/*
* Xilinx HW XABR Scaler filter with multipule resolution channels
*
* This file is part of FFmpeg.
*
* FFmpeg is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* FFmpeg is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include "vpi_types.h"
#include "vpi_api.h"

#include "avfilter.h"
#include "filters.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"
#include "libavutil/hwcontext.h"
#include "libavutil/hwcontext_vpe.h"
#if CONFIG_LIBXRM
#include "xrm_scale_interface.h"
#endif

#include "log_ama.h"

typedef enum {
    XABR_SESSION_ALL_RATE = 0,
    XABR_SESSION_FULL_RATE,
    XABR_MAX_SESSIONS
} XabrSessionType;

typedef enum {
    XABR_PSTATE_ENQUEUING = 0,
    XABR_PSTATE_STARTED,
    XABR_PSTATE_FLUSHING,
    XABR_PSTATE_CLOSED
} XabrPipelineState;

typedef struct VpeXabrScalerContext {
    const AVClass *class;
    AVBufferRef *hwdevice;
    AVBufferRef *hwframe;

#if CONFIG_LIBXRM
    XrmScaleContext xrm_ctx;
#endif

    VpiProcInitParams vpe_xabr_params;
    VpiProcHandle vpe_xabr_ctx[XABR_MAX_SESSIONS];

    /* options */
    int   nb_outputs;
    char  *str_resizes;
    char  *out_swfmt;
    char  *crop;
    char  *vpe_xabr_extra_params;

    VpiChannelInfo ch_info[VPE_CHANNEL_NUM];
    AVDictionary *dict;

    char        *out_rate_str;
    char        out_rate[VPE_CHANNEL_NUM][5];
    int         num_sessions;
    int         session_frame;
    AVRational  in_frame_rate;
    AVRational  out_frame_rate[VPE_CHANNEL_NUM];
    int         session_nb_outputs[XABR_MAX_SESSIONS];

    //Pipeline support
    int enable_pipeline;
    XabrPipelineState p_state[XABR_MAX_SESSIONS];
    int* inlink_copy;

    char instance_name[64];
} VpeXabrScalerContext;

static int xabr_out_config_props(AVFilterLink *outlink);
static int validate_rate_config(VpeXabrScalerContext *ctx);
static int get_num_scaler_sessions(VpeXabrScalerContext *ctx);
static int get_num_full_rate_outputs(VpeXabrScalerContext *ctx);
static void write_session_log(AVFilterLink *inlink);
static int vpe_xabr_filter_frame(AVFilterLink *inlink, AVFrame *inframe);

/**
 * get_avpixfmt_from_vpifmt: Retrieve value of AVformat from VPIformat
 *
 * @param format: Value of VPI format
 * @return Value of AVformat.
 */
static int get_avpixfmt_from_vpifmt(VpiFmt format)
{
    switch (format) {
        case VPI_FORMAT_YUV420P:        return AV_PIX_FMT_YUV420P;
        case VPI_FORMAT_NV12:           return AV_PIX_FMT_NV12;
        case VPI_FORMAT_YUV420P10BE:    return AV_PIX_FMT_YUV420P10BE;
        case VPI_FORMAT_YUV420P10LE:    return AV_PIX_FMT_YUV420P10LE;
        case VPI_FORMAT_P010BE:         return AV_PIX_FMT_P010BE;
        case VPI_FORMAT_P010LE:         return AV_PIX_FMT_P010LE;
        case VPI_FORMAT_SP1010:         return AV_PIX_FMT_PACKED10;
        case VPI_FORMAT_RGB24_P:        return AV_PIX_FMT_RGBP;

        default:
            return AV_PIX_FMT_NONE;
    }
}

/**
 * validate_rate_config: Validate Full & Half rate configuration
 *
 * @param VpeXabrScalerContext: The vpe scaler context
 * @return Success or Error
 */
static int validate_rate_config(VpeXabrScalerContext *ctx)
{
    int i, ret   = 0;
    int count_hr = 0;
    char *rate;

    //All outputs @half-rate not supported
    for (i = 0; i < ctx->nb_outputs; ++i) {
        rate = ctx->out_rate[i];
        if(strcmp(rate, "half") == 0) {
            count_hr += 1;
            ctx->out_frame_rate[i].num = ctx->in_frame_rate.num /2;
            ctx->out_frame_rate[i].den = ctx->in_frame_rate.den;
        } else if(strcmp(rate, "full") == 0) {
            ctx->out_frame_rate[i].num = ctx->in_frame_rate.num;
            ctx->out_frame_rate[i].den = ctx->in_frame_rate.den;
        } else {
          av_log(NULL, AV_LOG_ERROR, "output rate %s not supported\n", ctx->out_rate[i]);
          return AVERROR(EINVAL);
        }
    }
    ret = ((ctx->nb_outputs == count_hr) ? -1 : 0);

    if(ret ==-1) {
        av_log(NULL, AV_LOG_ERROR, "XABR:: All outputs at half-rate not supported\n");
    }

    return (ret);
}

/**
 * get_num_scaler_sessions: Get number scaler sessions which are
 *                          dependent on full/half rate config
 *
 * @param  VpeXabrScalerContext: The vpe scaler context
 * @return Number of scaler sessions
 */
static int get_num_scaler_sessions(VpeXabrScalerContext *ctx)
{
    int i;
    int count = 1;

    /* default = 1 session - full rate
       However if Mix out_rate is found then 2 sessions will
       be created to allow for frame drops.
    */
    for (i = 0; i < ctx->nb_outputs; ++i) {
        if(strcmp(ctx->out_rate[i], "full") != 0) {
            count = 2;
            break;
        }
    }
    return (count);
}

/**
 * get_num_full_rate_outputs: Get number of full rate outputs
 *
 * @param  VpeXabrScalerContext: The vpe scaler context
 * @return Number of full rate outputs or error
 */
static int get_num_full_rate_outputs(VpeXabrScalerContext *ctx)
{
    int i;
    int count = 0;
    int anchor_half_rate = 0;

    for (i = 0; i < ctx->nb_outputs; ++i) {
        if(strcmp(ctx->out_rate[i], "full") == 0) {
            if(anchor_half_rate) {
                av_log(NULL, AV_LOG_ERROR, "[%s][%d]ERROR : Full rate "
                       "specified after half rate! Full rate outputs must "
                       "preceed half rates. Output id %d\n", __func__, __LINE__, i);
                return AVERROR(EINVAL);
            }
            count += 1;
        } else {
            anchor_half_rate = 1;
        }
    }
    return (count);
}

static void set_instance_name(VpeXabrScalerContext *s, const char* ffmpeg_name)
{
    char *ptr = strrchr(ffmpeg_name, '_');
    snprintf(s->instance_name, 64, "scaler_ama%s", ptr);
    //printf("Name is %s\n", s->instance_name);
}

/**
 * write_session_log: Print Session logs
 *
 * @param  AVFilterLink: The AV Filter input link
 */
static void write_session_log(AVFilterLink *inlink)
{
    AVFilterContext *ctx    = inlink->dst;
    VpeXabrScalerContext *s = ctx->priv;
    int i, count;

    av_log(ctx, AV_LOG_DEBUG, "  XABR Session Configuration\n");
    av_log(ctx, AV_LOG_DEBUG, "---------------------------------------\n");
    av_log(ctx, AV_LOG_DEBUG, "Num Sessions = %d\n\n", s->num_sessions);

    for (count = 0; count < s->num_sessions; ++count) {
        av_log(ctx, AV_LOG_DEBUG, "Session:  %d\n", count);
        if(s->num_sessions > 1) {
            av_log(ctx, AV_LOG_DEBUG, "Type   :  %s\n", ((count) ? "FULL RATE ONLY" : "HALF RATE"));
        } else {
            av_log(ctx, AV_LOG_DEBUG, "Type   :  %s\n", "ALL RATE");
        }
        av_log(ctx, AV_LOG_DEBUG, "Num Out:  %d\n", s->session_nb_outputs[count]);
        for (i = 0; i < s->session_nb_outputs[count]; ++i) {
            av_log(ctx, AV_LOG_DEBUG, "out_%d :  (%4d x %4d) @%d fps\n", 
                   i, s->ch_info[i].width, s->ch_info[i].height, s->out_frame_rate[i].num);
        }
        av_log(ctx, AV_LOG_DEBUG, "--------------------------\n");
    }
}


/**
 * vpe_xabr_init: Initialize VPE ABR scaler parameters
 *
 * @param  AVFilterContext: The AV Filter context
 * @return Success or Error
 */
static av_cold int _vpe_xabr_init(AVFilterContext *ctx)
{
    VpeXabrScalerContext *s = ctx->priv;
    int i = 0;
    int ret = 0;
    char *p;
    char *tmp_str = NULL;
    char segments[VPE_CHANNEL_NUM][256];
    int seg_cnt;
    
    av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);
    av_log(ctx, AV_LOG_DEBUG, "ctx->hw_device_ctx = %p\n", ctx->hw_device_ctx);

    s->inlink_copy = NULL;

    /* process options and put options into dict */
    if(s->vpe_xabr_extra_params) {
        if((ret = av_dict_parse_string(&s->dict, s->vpe_xabr_extra_params, "=", "|", 0)) < 0) {
            return ret;
        }
    }

    if(s->enable_pipeline == 1){
        av_log(ctx, AV_LOG_WARNING, "Enabling pipeline in XABR\n");
    }

    s->vpe_xabr_params.enable_pipeline = s->enable_pipeline;
    if(s->enable_pipeline)
        s->p_state[0] = s->p_state[1] = XABR_PSTATE_ENQUEUING;
    else
        s->p_state[0] = s->p_state[1] = XABR_PSTATE_STARTED;

    if(s->crop){
        if((ret = av_dict_set(&s->dict, "crop_str", s->crop, 0)) < 0)
            return ret;
    }

    //set rate default to full
    for (i = 0; i < s->nb_outputs; i++) {
        strcpy(&s->out_rate[i][0], "full");
    }
    tmp_str = (char*)malloc(strlen(s->str_resizes)+1);
    if(!tmp_str)
        return AVERROR(ENOMEM);
    strcpy(tmp_str, s->str_resizes);
    seg_cnt=0;
    p = strtok(tmp_str, "()");
    while(p!=NULL){
        strcpy(&segments[seg_cnt++][0], p);
        p = strtok(NULL, "()");
    }

    for(i=0;i<seg_cnt;i++){
        p = strtok(&segments[i][0], "|");
        while(p != NULL){
            if (!strcmp(p, "full") || !strcmp(p, "half")){
                strcpy(&s->out_rate[i][0], p);
            }
            p = strtok(NULL, "|");
        }
    }
    free(tmp_str);

    for (i = 0; i < s->nb_outputs; i++) {
        char name[32];
        AVFilterPad pad = { 0 };

        snprintf(name, sizeof(name), "output%d", i);
        pad.type = AVMEDIA_TYPE_VIDEO;
        pad.name = av_strdup(name);
        if(!pad.name) {
            return AVERROR(ENOMEM);
        }

        pad.config_props = xabr_out_config_props;
        if((ret = ff_append_outpad_free_name(ctx, &pad)) < 0) {
            av_freep(&pad.name);
            return ret;
        }
    }
    av_log(s, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);
    return 0;
}

static av_cold int vpe_xabr_init(AVFilterContext *ctx)
{
    int ret;
    VpeXabrScalerContext *s = ctx->priv;
    set_instance_name(s, ctx->name);

    {
        LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ABR, s->instance_name);
        ret = _vpe_xabr_init(ctx);
        LOG_AMA_RESTORE_INSTANCE_INFO();
    }
    return ret;
}

/**
 * vpe_xabr_uninit: De-initialize VPE Scaler Plugin
 *
 * @param  AVFilterContext: The AV Filter context
 * @return Success or Error
 */
static av_cold void _vpe_xabr_uninit(AVFilterContext *ctx)
{
    int i;
    VpeXabrScalerContext *s = ctx->priv;

    av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);
#if CONFIG_LIBXRM
    xrm_scale_release(&s->xrm_ctx);
#endif
    for (i=0; i < s->num_sessions; ++i) {
        if(s->vpe_xabr_ctx[i] != NULL) {
            vpi_processor_close(s->vpe_xabr_ctx[i]);
        }
    }

    for (i = 0; i < s->nb_outputs; i++) {
        if(ctx->output_pads != NULL && ctx->output_pads[i].name != NULL) {
            av_freep(&ctx->output_pads[i].name);
        }
    }
    av_dict_free(&s->dict);
    av_buffer_unref(&s->hwframe);
    av_buffer_unref(&s->hwdevice);
}

static av_cold void vpe_xabr_uninit(AVFilterContext *ctx)
{
	VpeXabrScalerContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ABR, s->instance_name);
	_vpe_xabr_uninit(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
}

/**
 * vpe_xabr_init_hwdev: Initialize HW Device
 *
 * @param  AVFilterContext: The AV Filter context
 * @return Success or Error
 */
static int vpe_xabr_init_hwdev(AVFilterContext *ctx)
{
    VpeXabrScalerContext *s = ctx->priv;
    
    av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);

    if(!s->hwdevice) {
        if(ctx->hw_device_ctx) {
            s->hwdevice = av_buffer_ref(ctx->hw_device_ctx);
            if(!s->hwdevice) {
                return AVERROR(ENOMEM);
            }
        } else {
            return AVERROR_UNKNOWN;
        }
    }
    return 0;
}

/**
 * vpe_xabr_query_formats: Get Input/Output Pixel Format
 *
 * @param  AVFilterContext: The AV Filter context
 * @return Success or ERROR
 */
static int _vpe_xabr_query_formats(AVFilterContext *ctx)
{
    VpeXabrScalerContext *s = ctx->priv;
    AVVpeHWConfig *hwconfig = NULL;
    AVHWFramesConstraints *constraints = NULL;
    AVFilterFormats *in_fmts;
    AVFilterFormats *out_fmts;
    int ret;
    int i;
    
    av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);
    av_log(ctx, AV_LOG_DEBUG, "ctx->hw_device_ctx = %p\n", ctx->hw_device_ctx);
    
    ret = vpe_xabr_init_hwdev(ctx);
    if(ret < 0) {
        goto error;
    }
    
    hwconfig = av_hwdevice_hwconfig_alloc(s->hwdevice);
    if(!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }
    
    hwconfig->id = VPE_HW_XABR;
    av_log(ctx, AV_LOG_DEBUG, "hwconfig->id = %d\n", hwconfig->id);

    constraints =
        av_hwdevice_get_hwframe_constraints(s->hwdevice, hwconfig);
    if(!constraints) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    out_fmts = ff_make_format_list(constraints->valid_hw_formats);
    ret = ff_formats_ref(out_fmts, &ctx->outputs[0]->incfg.formats);
    if(ret < 0) {
        goto error;
    }

    in_fmts = ff_make_format_list(constraints->valid_hw_formats);
    i = 0;
    while (constraints->valid_sw_formats[i] != AV_PIX_FMT_NONE) {
        ret = ff_add_format(&in_fmts, constraints->valid_sw_formats[i]);
        if(ret < 0) {
            goto error;
        }
        i++;
    }
    ret = ff_formats_ref(in_fmts, &ctx->inputs[0]->outcfg.formats);
    if(ret < 0) {
        goto error;
    }

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);
    
    return 0;
    
error:
    av_log(ctx, AV_LOG_ERROR, "vpe_xabr_query_formats failed\n");
    
    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);
    
    return ret;
}

static int vpe_xabr_query_formats(AVFilterContext *ctx)
{
    int ret;
    VpeXabrScalerContext *s = ctx->priv;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ABR, s->instance_name);
	ret = _vpe_xabr_query_formats(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

/**
 * vpe_xabr_check_constraints: Check Hardware constraints
 *
 * @param  AVFilterLink: The AV Filter input link
 * @return Success or ERROR
 */
static int vpe_xabr_check_constraints(AVFilterLink *inlink)
{
    AVFilterContext *ctx = inlink->dst; 
    VpeXabrScalerContext *s = ctx->priv;
    AVVpeHWConfig *hwconfig = NULL;
    AVHWFramesConstraints *constraints = NULL;
    int ret;

    hwconfig = av_hwdevice_hwconfig_alloc(s->hwdevice);
    if(!hwconfig) {
        ret = AVERROR(ENOMEM);
        goto error;
    }
    
    hwconfig->id = VPE_HW_XABR;
    av_log(ctx, AV_LOG_DEBUG, "hwconfig->id = %d\n", hwconfig->id);

    constraints =
        av_hwdevice_get_hwframe_constraints(s->hwdevice, hwconfig);
    if(!constraints) {
        ret = AVERROR(ENOMEM);
        goto error;
    }

    av_log(ctx, AV_LOG_DEBUG, "get constraints %d,%d,%d,%d\n", constraints->min_width,
                    constraints->min_height, constraints->max_width, constraints->max_height);
    
    if(inlink->w < constraints->min_width  ||
       inlink->h < constraints->min_height ||
       inlink->w > constraints->max_width  ||
       inlink->h > constraints->max_height) {
        av_log(ctx, AV_LOG_ERROR, "Hardware does not support image "
               "size %dx%d (constraints: width %d-%d height %d-%d).\n",
               inlink->w, inlink->h,
               constraints->min_width,  constraints->max_width,
               constraints->min_height, constraints->max_height);
        ret = AVERROR(EINVAL);
        goto error;
    }

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);
    
    return 0;
    
error:
    av_log(ctx, AV_LOG_ERROR, "vpe_xabr_check_constraints failed\n");
    
    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);
    
    return ret;

}

/**
 * vpe_xabr_init_hwctx: Get HW Channel Information
 *
 * @param  AVFilterLink: The AV Filter input link
 * @return Success or ERROR
 */
static int vpe_xabr_init_hwctx(AVFilterLink *inlink)
{
    AVFilterContext *ctx = inlink->dst; 
    VpeXabrScalerContext *s = ctx->priv;
    VpiProcInitParams *param = &s->vpe_xabr_params;
    AVHWFramesContext *hwframe_ctx;
    AVHWDeviceContext *device_ctx;
    AVVpeDeviceContext *device_hwctx;
    AVVpeFramesContext *frame_hwctx;
    int ret, ch_index;

    /* hw device & frame init */
    if(inlink->hw_frames_ctx) {
        s->hwframe = av_buffer_ref(inlink->hw_frames_ctx);
        if(!s->hwframe) {
            return AVERROR(ENOMEM);
        }

        hwframe_ctx = (AVHWFramesContext*)s->hwframe->data;
    } else {
        av_log(ctx, AV_LOG_ERROR, "vpe_xabr can't get hwframe\n");
        return AVERROR_UNKNOWN;
    }

    ret = vpe_xabr_check_constraints(inlink);
    if(ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpe_xabr_check_constraints failed\n");
        return ret;
    }

    device_ctx   = hwframe_ctx->device_ctx;
    device_hwctx = device_ctx->hwctx;
    frame_hwctx  = hwframe_ctx->hwctx;

    param->dev_ctx = device_hwctx->vpe_dev_ctx;
    param->mem_ctx = frame_hwctx->vpe_mem_ctx;

    ret = vpi_memory_get_channel_info(frame_hwctx->vpe_mem_ctx, &frame_hwctx->proc_info->ch_info[0]);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_get_info failed\n");
        return AVERROR_EXTERNAL;
    }

    ret = vpi_memory_get_map_info(frame_hwctx->vpe_mem_ctx, 0, &ch_index);
    if (ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_get_map_info failed\n");
        return AVERROR_EXTERNAL;
    }

    param->walign = frame_hwctx->proc_info->ch_info[ch_index].walign;
    param->halign = frame_hwctx->proc_info->ch_info[ch_index].halign;
    param->flag   = frame_hwctx->proc_info->ch_info[ch_index].flag;
    av_log(ctx, AV_LOG_DEBUG, "ch_index = %d\n", ch_index);
    param->channel_choose = ch_index; /* add channel choose */

    return 0;
}


/**
 * xabr_out_config_props: Initialize Scaler Output Properties
 * this function is called in reverse chronological order of active output
 * @param  AVFilterLink: The AV Filter input link
 * @return Success or ERROR
 */
static int _xabr_out_config_props(AVFilterLink *outlink)
{
    AVFilterContext      *ctx = outlink->src;
    VpeXabrScalerContext *s   = ctx->priv;
    const int outlink_idx     = FF_OUTLINK_IDX(outlink);
    AVFilterLink       *out   = ctx->outputs[outlink_idx]; //select output link at given index
    AVHWFramesContext *hwframe;
    AVVpeFramesContext *vpe_frame_hwctx;
    int ret;

    av_log(ctx, AV_LOG_DEBUG, "%s(%d)\n", __FUNCTION__, __LINE__);

    /* hw device & frame init */
    if(out->hw_frames_ctx) {
        //not expecting the out to have hwframe ctx
        av_log(ctx, AV_LOG_DEBUG, "vpe xabr output %d hwframe ctx exists\n", outlink_idx);
        goto error;
    }
    out->hw_frames_ctx = av_hwframe_ctx_alloc(s->hwdevice);
    if(!out->hw_frames_ctx) {
        av_log(ctx, AV_LOG_ERROR, "vpe xabr output_%d hw_frames_ctx alloc failed\n",outlink_idx);
        ret = AVERROR(ENOMEM);
        goto error;
    }

    hwframe = (AVHWFramesContext*)out->hw_frames_ctx->data;

    if(!hwframe->pool) {
        av_log(ctx, AV_LOG_DEBUG, "vpe xabr init output %d hwframe\n", outlink_idx);
        hwframe->width      = s->ch_info[outlink_idx].width;
        hwframe->height     = s->ch_info[outlink_idx].height;
        hwframe->format     = AV_PIX_FMT_VPE;
        hwframe->sw_format  = get_avpixfmt_from_vpifmt(s->ch_info[outlink_idx].sw_format);
        if((ret = av_hwframe_ctx_init(out->hw_frames_ctx)) < 0) {
            av_log(ctx, AV_LOG_ERROR, "av_hwframe_ctx_init for output %d failed\n", outlink_idx);
            goto error;
        }
    } else {
        av_log(ctx, AV_LOG_DEBUG, "vpe xabr output %d hwframe pool exists\n", outlink_idx);
        goto error;
    }

    vpe_frame_hwctx = hwframe->hwctx;

    av_log(ctx, AV_LOG_DEBUG, "%s(%d) output hwframe @%p\n",
        __FUNCTION__, __LINE__, vpe_frame_hwctx);

    out->w  = s->ch_info[outlink_idx].width;
    out->h  = s->ch_info[outlink_idx].height;

    //Set out fps for each channel
    outlink->frame_rate.num = s->out_frame_rate[outlink_idx].num;
    outlink->frame_rate.den = s->out_frame_rate[outlink_idx].den;

    //set frame_hwcontext map and ch_info
    //vpe_frame_hwctx->vpe_ch_map[outlink_idx] = outlink_idx;
    memcpy(&vpe_frame_hwctx->proc_info->ch_info[outlink_idx], &s->ch_info[outlink_idx], sizeof(VpiChannelInfo));

    //set memory context map and ch_info
    ret = vpi_memory_set_map_info(vpe_frame_hwctx->vpe_mem_ctx, 0, 0);
    if(ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_set_map_info failed\n");
        goto error;
    }
    ret = vpi_memory_set_single_channel_info(vpe_frame_hwctx->vpe_mem_ctx, 0, &s->ch_info[outlink_idx], VPE_HW_XABR);
    if(ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_memory_set_single_channel_info failed\n");
        goto error;
    }
    return 0;

error:
    av_log(ctx, AV_LOG_ERROR, "xabr_out_config_props failed\n");
    return ret;
}

static int xabr_out_config_props(AVFilterLink *outlink)
{
	int ret;
	AVFilterContext      *ctx = outlink->src;
    VpeXabrScalerContext *s = ctx->priv;

    {
        LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ABR, s->instance_name);
        ret = _xabr_out_config_props(outlink);
        LOG_AMA_RESTORE_INSTANCE_INFO();
    }
    return ret;
}

/**
 * vpe_xabr_config_props: Initialize ffmpeg Scaler Plugin
 * @param  AVFilterLink: The AV Filter input link
 * @return Success or ERROR
 */
static int _vpe_xabr_config_props(AVFilterLink *inlink)
{
    AVFilterContext *ctx = inlink->dst; 
    VpeXabrScalerContext *s = ctx->priv;
    VpiProcInitParams *p_params = &s->vpe_xabr_params;
    VpiProcInfo info;
    AVDictionaryEntry *en = NULL;
    int count, ret = 0;
#if CONFIG_LIBXRM
    XrmInterfaceProperties in_xrm_props;
    XrmInterfaceProperties* out_xrm_props;
    int dev_index = -1;
    int props_index = 0;
    memset(&in_xrm_props, 0, sizeof(in_xrm_props));
#endif

    ret = vpe_xabr_init_hwctx(inlink);
    if(ret < 0) {
        return ret;
    }

    //get input frame rate
    if(inlink->time_base.den > 0) {
        s->in_frame_rate.num = inlink->frame_rate.num;
        s->in_frame_rate.den = inlink->frame_rate.den;
    }

    //validate rate configuration params
    ret = validate_rate_config(s);
    if(ret < 0) {
        return AVERROR(EINVAL);
    }

    //determine number of sessions to create
    s->num_sessions = get_num_scaler_sessions(s);

    //All-rate session includes all outputs
    s->session_nb_outputs[XABR_SESSION_ALL_RATE] = s->nb_outputs;

     if(s->num_sessions > 1) {
        s->session_nb_outputs[XABR_SESSION_FULL_RATE] = get_num_full_rate_outputs(s);
        if(s->session_nb_outputs[XABR_SESSION_FULL_RATE] < 0) {
            av_log(ctx, AV_LOG_ERROR, "Must have at-least 1 full rate session\n");
            return AVERROR(EINVAL);
        }
    }

    p_params->processor_type = VPE_PROCESSOR_XABR;
    p_params->width          = inlink->w;
    p_params->height         = inlink->h;
    p_params->input_format   = (char *)av_get_pix_fmt_name(inlink->format);
    if(inlink->format == AV_PIX_FMT_VPE) {
        AVHWFramesContext *fmctx = (AVHWFramesContext *)inlink->hw_frames_ctx->data;
        p_params->input_sw_format = (char *)av_get_pix_fmt_name(fmctx->sw_format);
    } else {
        p_params->input_sw_format = (char *)av_get_pix_fmt_name(inlink->format);
    }
    p_params->resize_str = s->str_resizes;


    p_params->mixrate_ctx = NULL;
    //allocte the scaler context for all sessions
    for (count = 0; count < s->num_sessions; ++count) {
        if(count){
            p_params->mixrate_ctx = s->vpe_xabr_ctx[count-1];
        }
        p_params->output_num = s->session_nb_outputs[count];

        ret = vpi_processor_init(p_params, &s->vpe_xabr_ctx[count]);
        if(ret < 0) {
            av_log(ctx, AV_LOG_ERROR, "VPE processor scaler init error: %s\n", vpi_error_str(ret));
            return AVERROR_EXTERNAL;
        }

        while((en = av_dict_get(s->dict, "", en, AV_DICT_IGNORE_SUFFIX))) {
            if(ret = vpi_processor_set_param(s->vpe_xabr_ctx[count], en->key, en->value))
                av_log(ctx, AV_LOG_WARNING,
                       "Error parsing option '%s = %s'. Please check if it is valid for xabr options. error = %s.\n",
                       en->key, en->value, vpi_error_str(ret));
        }

        //initialize scaler
        ret = vpi_processor_open(s->vpe_xabr_ctx[count]);
        if(ret < 0) {
            av_log(ctx, AV_LOG_ERROR, "vpi_processor_open failed, error=%s\n", vpi_error_str(ret));
            return AVERROR_EXTERNAL;
        }
    }
    //extract output resolutions (channel info) from xabr vpe plugin, Only for ALL_RATE session
    ret = vpi_processor_get_info(s->vpe_xabr_ctx[0], &info);
    if(ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "vpi_processor_get_info failed, error=%s\n", vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }
    memcpy(s->ch_info, &info.ch_info, sizeof(VpiChannelInfo)*VPE_CHANNEL_NUM);

#if CONFIG_LIBXRM
    // Reserve xrm resource
    in_xrm_props.width = s->vpe_xabr_params.width;
    in_xrm_props.height = s->vpe_xabr_params.height;
    in_xrm_props.fps_num = s->in_frame_rate.num;
    in_xrm_props.fps_den = s->in_frame_rate.den;

    out_xrm_props = calloc(1, ctx->nb_outputs * sizeof(XrmInterfaceProperties));
    for (count = 0; count < ctx->nb_outputs; count++) {
        if (!s->ch_info[count].enabled) {
            continue;
        }
        out_xrm_props[props_index].width   = s->ch_info[count].width;
        out_xrm_props[props_index].height  = s->ch_info[count].height;
        out_xrm_props[props_index].fps_num = s->out_frame_rate[count].num;
        out_xrm_props[props_index].fps_den = s->out_frame_rate[count].den;
        props_index++;
    }

    if(vpi_device_get_device_index(s->vpe_xabr_params.dev_ctx, &dev_index) < 0) {
        return AVERROR(EINVAL);
    }
    if (xrm_scale_reserve(&s->xrm_ctx, dev_index, &in_xrm_props, out_xrm_props, props_index) == XRM_ERROR) {
        vpe_xabr_uninit(ctx);
        return AVERROR(EBUSY);
    }
	free(out_xrm_props);
#endif
    s->session_frame = 0; //start with even frame
    //log session configuration
    write_session_log(inlink);

    return 0;
}

static int vpe_xabr_config_props(AVFilterLink *inlink)
{
    int ret;
    AVFilterContext *ctx = inlink->dst;
    VpeXabrScalerContext *s = ctx->priv;

    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ABR, s->instance_name);
    ret = _vpe_xabr_config_props(inlink);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}


static void vpe_xabr_filter_flush(AVFilterLink *link)
{
    AVFilterContext *ctx = link->dst;
    VpeXabrScalerContext *s = ctx->priv;
    AVHWFramesContext *hwframe_ctx;
    AVVpeFramesContext *frame_hwctx;
    VpiFrm *nFrm;
    AVFrame *nframe;
    int ret = -1;
    int *inlink = (int *)link;
    int sessions_flushed = 0;

    if(s->enable_pipeline == 0){
        return;
    }

    nframe = av_frame_alloc();
    nframe->format = AV_PIX_FMT_VPE;

    hwframe_ctx = (AVHWFramesContext*)s->hwframe->data;
    frame_hwctx = hwframe_ctx->hwctx;

    ret = vpi_frame_alloc(frame_hwctx->vpe_mem_ctx, &nFrm);
    if(ret < 0) {
        av_log(ctx, AV_LOG_ERROR, "failed to create dummy Vpi frame\n");
        return;
    }
    nframe->data[0] = (uint8_t*)nFrm;

    for(int i=0;i < s->num_sessions; i++){
        s->p_state[i] = XABR_PSTATE_FLUSHING;
    }

    //For pipeline depth >1 it is necessary to flush both scaler session one by one.
    while(sessions_flushed != s->num_sessions){
        for(int i=0;i < s->num_sessions; i++){
            if(inlink == s->inlink_copy){
                nFrm->vpi_picinfo_buf = NULL;
                if(s->p_state[i] == XABR_PSTATE_FLUSHING){
                    av_log(ctx, AV_LOG_VERBOSE, "Flushing a frame from %s session\n", i == 0 ? "ALL_RATE" : "FULL_RATE");
                    ret = vpe_xabr_filter_frame(link, nframe);
                    if(ret < 0){
                        av_log(ctx, AV_LOG_ERROR, "filter frame error while flushing\n");
                        sessions_flushed++;
                        s->p_state[i] = XABR_PSTATE_CLOSED;
                        continue; 
                    }
                }
                else{
                    sessions_flushed++;
                    av_log(ctx, AV_LOG_VERBOSE, "flushing of %s session complete\n", i == 0 ? "ALL_RATE" : "FULL_RATE");
                }
            }
        }
    }
    
    vpi_frame_free(frame_hwctx->vpe_mem_ctx, nFrm);
    av_frame_free(&nframe);
}

/**
 * vpe_xabr_filter_frame:  Process filter frame
 * @param  AVFilterLink: The AV Filter input link
 * @param  AVFrame: Input filter frame
 * @return Success or ERROR
 */
static int vpe_xabr_filter_frame(AVFilterLink *inlink, AVFrame *inframe)
{
    AVFilterContext *ctx = inlink->dst;
    VpeXabrScalerContext *s = ctx->priv;
    int ret, i;
    VpiFrm *in_pic    = NULL;
    VpiFrm *out_vpiframes[VPE_CHANNEL_NUM] = {0};
    AVFrame *out_avframes[VPE_CHANNEL_NUM] = {0};
    int session_num_out = 0;
    VpiProcHandle   *curr_ctx;
    XabrSessionType session_type;
    AVFrame *tmp;

    s->inlink_copy = (int*)inlink;

    if(inframe->format == AV_PIX_FMT_VPE) {
        in_pic = (VpiFrm *)inframe->data[0];
    } else {
        return AVERROR_UNKNOWN;
    }

    if(s->num_sessions > 1) {
        //Even Frame = SC_SESSION_ALL_RATE Odd Frame = SC_SESSION_FULL_RATE
        session_type = ((s->session_frame & 0x01) ? XABR_SESSION_FULL_RATE : XABR_SESSION_ALL_RATE);
        s->session_frame = (s->session_frame + 1) % XABR_MAX_SESSIONS;
    } else {
        session_type = XABR_SESSION_ALL_RATE;
    }

    curr_ctx        = s->vpe_xabr_ctx[session_type];
    session_num_out = s->session_nb_outputs[session_type];


    if(s->p_state[session_type] != XABR_PSTATE_ENQUEUING){
        for(i=0; i<session_num_out; i++){
            out_avframes[i] = av_frame_alloc();
            if(out_avframes[i] == NULL){
                av_log(ctx, AV_LOG_ERROR, "av_frame_alloc fail for out_avframes[%d] ret = %d\n", i, ret);
                goto error_exit;
            }

            ret = av_hwframe_get_buffer(ctx->outputs[i]->hw_frames_ctx, out_avframes[i], 0);
            if(ret < 0){
                av_log(ctx, AV_LOG_ERROR, "av_hwframe_get_buffer fail for out_avframes[%d] ret = %d\n", i, ret);
                goto error_exit;
            }

            out_vpiframes[i] = (VpiFrm*)out_avframes[i]->data[0];

            ret = av_frame_copy_props(out_avframes[i], inframe);
            if(ret < 0){
                av_log(ctx, AV_LOG_ERROR, "av_frame_copy_props fail for out_avframes[%d] ret = %d\n", i, ret);
                goto error_exit;
            }
        }
    }

    in_pic->opaque = (void*)inframe;
    in_pic->pts = inframe->pts;
    ret = vpi_processor_process_frame_list(curr_ctx, in_pic, out_vpiframes);
    switch(ret){
        case VPI_SUCCESS:
            if(s->p_state[session_type] == XABR_PSTATE_FLUSHING){
                s->p_state[session_type] = XABR_PSTATE_CLOSED;
                av_log(ctx, AV_LOG_VERBOSE, "flushing done p_state(%d) changed to %d\n", session_type, s->p_state[session_type]);
            }
            break;
        case XABR_SEND_MORE_DATA:
            av_log(ctx, AV_LOG_VERBOSE, "vpi_processor_process_frame needs more frames(%d)\n", session_type);
            return 0;
        case XABR_PIPELINE_FILL_DONE:
            av_log(ctx, AV_LOG_VERBOSE, "We start recieving outputs from next frame(%d)\n", session_type);
            s->p_state[session_type] = XABR_PSTATE_STARTED;
            return 0;
        case XABR_FLUSH_AGAIN:
            av_log(ctx, AV_LOG_VERBOSE, "Continue frame flush (session type = %d)\n", session_type);
            break;
        default:
            av_log(ctx, AV_LOG_ERROR, "vpi_processor_process_frame failed, error=%s\n", vpi_error_str(ret));
            goto error_exit;
            break;
    }

    for (i = 0; i < session_num_out; i++) {
        int *index;

        out_avframes[i]->buf[1] = av_buffer_alloc(sizeof(int));
        index = (int*)out_avframes[i]->buf[1]->data;
        *index = 0; // Used in hwcontext_vpe as output channel index

        out_avframes[i]->format = AV_PIX_FMT_VPE;
        out_avframes[i]->width = out_vpiframes[i]->ch_info[0].width;
        out_avframes[i]->height = out_vpiframes[i]->ch_info[0].height;
        out_avframes[i]->pts = out_vpiframes[i]->pts;
        av_log(ctx, AV_LOG_TRACE, "[%p]session_type=%d out_pts = %ld\n",&out_vpiframes[i], session_type, out_avframes[i]->pts);
        ret = ff_filter_frame(ctx->outputs[i], out_avframes[i]);

        if(ret < 0) {
            av_log(ctx, AV_LOG_ERROR, "filter frame failed, error=%d\n", ret);
            goto error_exit;
        }
    }

    tmp = (AVFrame*)out_vpiframes[0]->opaque;
    av_frame_free(&tmp);

    return 0;

error_exit:
    for(i=0;i<session_num_out;i++){
        if(out_avframes[i])
            av_frame_unref(out_avframes[i]);
    }

    return ret;
}

static int _vpe_xabr_activate(AVFilterContext *ctx)
{
    AVFrame *frame = NULL;
    int ret, status;
    int64_t pts;

    for(int i=0;i<ctx->nb_outputs;i++)
        FF_FILTER_FORWARD_STATUS_BACK(ctx->outputs[i], ctx->inputs[0]);

    ret = ff_inlink_consume_frame(ctx->inputs[0], &frame);
    if (ret < 0)
        return ret;

    if (frame) {
        int ret =  vpe_xabr_filter_frame(ctx->inputs[0], frame);
        if(ret < 0 )
            return ret;
        ff_filter_set_ready(ctx, 10);
        return ret;
    }

    ret = ff_inlink_acknowledge_status(ctx->inputs[0], &status, &pts);
    if (ret > 0) {
        /* flush */
        vpe_xabr_filter_flush(ctx->inputs[0]);
        for(int i=0;i<ctx->nb_outputs;i++)
            ff_outlink_set_status(ctx->outputs[i], status, pts);
        return 0;
    }
    if(!ret){
        for(int i=0;i<ctx->nb_outputs;i++)
            FF_FILTER_FORWARD_WANTED(ctx->outputs[i], ctx->inputs[0]);
        return 0;
    }

    return FFERROR_NOT_READY;
}

static int vpe_xabr_activate(AVFilterContext *ctx)
{
    int ret;
    VpeXabrScalerContext *s = ctx->priv;
    LOG_AMA_SET_AND_SAVE_INSTANCE_INFO(LOG_AMA_ACCEL_ABR, s->instance_name);
    ret = _vpe_xabr_activate(ctx);
    LOG_AMA_RESTORE_INSTANCE_INFO();
    return ret;
}

#define OFFSET(x) offsetof(VpeXabrScalerContext, x)
#define FLAGS (AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_FILTERING_PARAM)
static const AVOption vpe_xabr_options[] = {
    { "outputs", "set number of outputs", OFFSET(nb_outputs), AV_OPT_TYPE_INT, { .i64 = 4 }, 1, VPE_CHANNEL_NUM, FLAGS },
    { "out_res", "set resize configuration", OFFSET(str_resizes), AV_OPT_TYPE_STRING, {.str=NULL}, .flags = FLAGS},
    { "crop", "set scrop for first output channel", OFFSET(crop), AV_OPT_TYPE_STRING, {.str=NULL}, .flags = FLAGS},
    { "expert_options", "set xabr extra params", OFFSET(vpe_xabr_extra_params), AV_OPT_TYPE_STRING, {.str=NULL}, .flags = FLAGS},
    { "enable_pipeline", "Enable pipelining for XABR scaler", OFFSET(enable_pipeline), AV_OPT_TYPE_INT, { .i64 = 1 }, 0, 1, FLAGS },
    { NULL }
};


AVFILTER_DEFINE_CLASS(vpe_xabr);

static const AVFilterPad vpe_xabr_inputs[] = {
    {
        .name         = "default",
        .type         = AVMEDIA_TYPE_VIDEO,
        .config_props = vpe_xabr_config_props,
    }
};

AVFilter ff_vf_vpe_xabr = {
    .name           = "scaler_ama",
    .description    = NULL_IF_CONFIG_SMALL("AMD-Xilinx ABR Scaler."),
    .priv_size      = sizeof(VpeXabrScalerContext),
    .priv_class     = &vpe_xabr_class,
    .init           = vpe_xabr_init,
    .uninit         = vpe_xabr_uninit,
    .nb_outputs     = 0,
    FILTER_QUERY_FUNC(vpe_xabr_query_formats),
    FILTER_INPUTS(vpe_xabr_inputs),
    .activate       = vpe_xabr_activate,
    .flags          = AVFILTER_FLAG_DYNAMIC_OUTPUTS,
    .flags_internal = FF_FILTER_FLAG_HWFRAME_AWARE,
};
